/*
 * PhonicEQ Inc.  E100P/T100P RAVEN E1/PRA card Driver
 *
 * Written by Marcin Pycko <mpycko@phoniceq.com>
 *
 * Derived from the code written by Mark Spencer <markster@digium.com>
 *            Matthew Fredrickson <creslin@digium.com>
 *            William Meadows <wmeadows@digium.com>
 *
 * Copyright (C) 2006, PhonicEQ, Inc.
 * Copyright (C) 2001, Linux Support Services, Inc.
 *
 * All rights reserved.
 *
 */

/*
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2 as published by the
 * Free Software Foundation. See the LICENSE file included with
 * this program for more details.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/usb.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#ifdef STANDALONE_ZAPATA
#include "zaptel.h"
#else
#include <zaptel/zaptel.h>
#endif
#ifdef LINUX26
#include <linux/moduleparam.h>
#endif

#define WC_MAX_CARDS	32

/* comment this out if you don't use the data mode */
#define DATALINK

/* uncomment this if your circuit dies sometimes */
/* #define SANITY_CHECK */

/* uncomment this to select E1 mode as default */
/* #define DEFAULT_TYPE_E1  1 */


/*
#define TEST_REGS
*/

/* Define to get more attention-grabbing but slightly more I/O using
   alarm status */
#define FANCY_ALARM


#define DELAY	0x0	/* 30 = 15 cycles, 10 = 8 cycles, 0 = 3 cycles */

#define WC_CNTL    	0x00
#define WC_OPER		0x01
#define WC_AUXC    	0x02
#define WC_AUXD    	0x03
#define WC_MASK0   	0x04
#define WC_MASK1   	0x05
#define WC_INTSTAT 	0x06

#define WC_DMAWS	0x08
#define WC_DMAWI	0x0c
#define WC_DMAWE	0x10
#define WC_DMARS	0x18
#define WC_DMARI	0x1c
#define WC_DMARE	0x20
#define WC_CURPOS	0x24

#define WC_SERC		0x2d
#define WC_FSCDELAY	0x2f

#define WC_USERREG	0xc0

#define WC_CLOCK	0x0
#define WC_LEDTEST	0x1
#define WC_VERSION	0x2

/* Offset between transmit and receive */
#define WC_OFFSET	4

#define BIT_CS		(1 << 7)
#define BIT_ADDR	(0xf)

#define BIT_LED0	(1 << 1)
#define BIT_LED1	(1 << 0)

#define TYPE_T1		0
#define TYPE_E1		1

static char *chips[] =
{
	"DS2152",
	"DS21352",
	"DS21552",
	"Unknown Chip (3)",
	"DS2154",
	"DS21354",
	"DS21554",
	"Unknown Chip (7)",
};

static int chanmap_t1[] = 
{
0, 7, 6, 5, 11, 10, 9, 15, 14, 13, 19, 18, 17, 23, 22, 21, 27, 26, 25, 31, 30, 29, 3, 2, 1, 
};

static int chanmap_e1[] = 
{
0, 7, 6, 5, 4, 11, 10, 9, 8, 15, 14, 13, 12, 19, 18, 17, 16, 23, 22, 21, 20, 27, 26, 25, 24, 31, 30, 29, 28, 3, 2, 1, 
};

#ifdef FANCY_ALARM
static int altab[] = {
0, 0, 0, 1, 2, 3, 4, 6, 8, 9, 11, 13, 16, 18, 20, 22, 24, 25, 27, 28, 29, 30, 31, 31, 32, 31, 31, 30, 29, 28, 27, 25, 23, 22, 20, 18, 16, 13, 11, 9, 8, 6, 4, 3, 2, 1, 0, 0, 
};
#endif

struct t1xxp {
	struct pci_dev *dev;
	spinlock_t lock;
	int ise1;
	int num;
	char *variety;
	unsigned int intcount;
	int usecount;
	int clocktimeout;
	int sync;
	int dead;
	int blinktimer;
	int alarmtimer;
	int loopupcnt;
	int loopdowncnt;
	int miss;
	int misslast;
	int *chanmap;
	int chanfound;
#ifdef FANCY_ALARM
	int alarmpos;
#endif
	unsigned char ledtestreg;
	unsigned char outbyte;
	unsigned long ioaddr;
	unsigned short canary;
	/* RBS signalling */
	unsigned char txsigs[16];
	dma_addr_t 	readdma;
	dma_addr_t	writedma;
	volatile unsigned char *writechunk;					/* Double-word aligned write memory */
	volatile unsigned char *readchunk;					/* Double-word aligned read memory */
	unsigned char ec_chunk1[31][ZT_CHUNKSIZE];
	unsigned char ec_chunk2[31][ZT_CHUNKSIZE];
	unsigned char tempo[32];
	unsigned char temporx[32];
	struct zt_span span;						/* Span */
	struct zt_chan chans[31];					/* Channels */
};

#define CANARY 0xca1e

int debug = 0;   /* doesnt do anything */
#ifdef DEFAULT_TYPE_E1
int mode = TYPE_E1;
#else
int mode = TYPE_T1;
#endif
int japan = 0;

static struct t1xxp *cards[WC_MAX_CARDS];

static inline void wait(unsigned int ms)
{
	unsigned long endjiffies;
	endjiffies = jiffies + ms;
	while(endjiffies < jiffies);
}

static inline void start_alarm(struct t1xxp *wc)
{
#ifdef FANCY_ALARM
	wc->alarmpos = 0;
#endif
	wc->blinktimer = 0;
}

static inline void stop_alarm(struct t1xxp *wc)
{
#ifdef FANCY_ALARM
	wc->alarmpos = 0;
#endif
	wc->blinktimer = 0;
}

static inline void __select_framer(struct t1xxp *wc, int reg)
{
	/* Top four bits of address from AUX 6-3 */
	wc->outbyte &= ~BIT_CS;
	wc->outbyte &= ~BIT_ADDR;
	wc->outbyte |= (reg & 0xf0) >> 4;
	outb(wc->outbyte, wc->ioaddr + WC_AUXD);
}

static inline void __select_control(struct t1xxp *wc)
{
	if (!(wc->outbyte & BIT_CS)) {
		wc->outbyte |= BIT_CS;
		outb(wc->outbyte, wc->ioaddr + WC_AUXD);
	}
}

static int t1xxp_open(struct zt_chan *chan)
{
	struct t1xxp *wc = chan->pvt;
	if (wc->dead)
		return -ENODEV;
	wc->usecount++;
#ifndef LINUX26	
	MOD_INC_USE_COUNT;
#endif	
	return 0;
}

static int __t1_get_reg(struct t1xxp *wc, int reg)
{
	unsigned char res;
	__select_framer(wc, reg);
	/* Get value */
	res = inb(wc->ioaddr + WC_USERREG + ((reg & 0xf) << 2));
	return res;
}

static int __t1_set_reg(struct t1xxp *wc, int reg, unsigned char val)
{
	__select_framer(wc, reg);
	/* Send address */
	outb(val, wc->ioaddr + WC_USERREG + ((reg & 0xf) << 2));
	return 0;
}

static void t1xxp_release(struct t1xxp *wc)
{
	zt_unregister(&wc->span);
	kfree(wc);
	printk("Freed the card\n");
}

static int t1xxp_close(struct zt_chan *chan)
{
	struct t1xxp *wc = chan->pvt;
	wc->usecount--;
#ifndef LINUX26	
	MOD_DEC_USE_COUNT;
#endif
	/* If we're dead, release us now */
	if (!wc->usecount && wc->dead) 
		t1xxp_release(wc);
	return 0;
}

static void t1xxp_enable_interrupts(struct t1xxp *wc)
{
	/* Clear interrupts */
	outb(0xff, wc->ioaddr + WC_INTSTAT);
	/* Enable interrupts (we care about all of them) */
	outb(0x3c /* 0x3f */, wc->ioaddr + WC_MASK0); 
	/* No external interrupts */
	outb(0x00, wc->ioaddr + WC_MASK1);
}

static void t1xxp_restart_dma(struct t1xxp *wc)
{
	/* Reset Master and serial */
	outb(0x01, wc->ioaddr + WC_CNTL);
	outb(0x01, wc->ioaddr + WC_OPER);
}

static int t1xxp_watchdog(struct zt_span *span, int event)
{
	printk("Restarting DMA\n");
	t1xxp_restart_dma(span->pvt);
	return 0;
}

static void t1xxp_start_dma(struct t1xxp *wc)
{
	/* Reset Master and TDM */
	outb(DELAY | 0x0f, wc->ioaddr + WC_CNTL);
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(1);
	outb(DELAY | 0x01, wc->ioaddr + WC_CNTL);

	outb((1<<4), wc->ioaddr + WC_AUXD);

	outb(0x01, wc->ioaddr + WC_OPER);
	outb(0x03, wc->ioaddr + WC_OPER);

	/* Wait 100ms to give plenty of time for reset */
	wait(100);

	outb(0x01, wc->ioaddr + WC_OPER);

	/* Wait 100ms to give plenty of time for reset */
	wait(100);

	outb(0x0, wc->ioaddr + WC_AUXD);

	if (debug) printk("Started DMA\n");
}

static void __t1xxp_stop_dma(struct t1xxp *wc)
{
	outb(0x00, wc->ioaddr + WC_OPER);
}

static void __t1xxp_disable_interrupts(struct t1xxp *wc)	
{
	outb(0x00, wc->ioaddr + WC_MASK0);
	outb(0x00, wc->ioaddr + WC_MASK1);
}

static void __t1xxp_set_clear_t1(struct t1xxp *wc)
{
	int i,j;
	unsigned short val=0;
	for (i = 0; i < 24; i++) {
		j = (i/8);
		if (!(wc->span.chans[i].flags & ZT_FLAG_CLEAR))
			val |= 1 << (i % 8);

		if ((i % 8)==7) {
			__t1_set_reg(wc, 0x8 + j, val);
			val = 0;
		}
	}
}

static void __t1xxp_set_clear_e1(struct t1xxp *wc)
{
	int i,j;
	unsigned short val;
	if (wc->span.lineconfig & ZT_CONFIG_CCS)
		return;
	val = 1;
	for (i = 1; i < 16; i++) {
		j = (i/8);
		if (!(wc->span.chans[i].flags & ZT_FLAG_CLEAR))
			val |= 1 << (i % 8);
		if ((i % 8)==7) {
			__t1_set_reg(wc, 0x8 + j, val);
			val = 0;
		}
	}
	val = 1;
	for (i = 17; i < 32; i++) {
		j = (i/8);
		if (!(wc->span.chans[i].flags & ZT_FLAG_CLEAR))
			val |= 1 << (i % 8);
		if ((i % 8)==7) {
			__t1_set_reg(wc, 0x8 + j, val);
			val = 0;
		}
	}
}

static void t1xxp_t1_framer_start(struct t1xxp *wc)
{
	char *coding, *framing;
	int alreadyrunning = wc->span.flags & ZT_FLAG_RUNNING;
	unsigned long flags;
	int lc = wc->span.lineconfig;

	spin_lock_irqsave(&wc->lock, flags);

	if (!alreadyrunning) {
		__t1xxp_set_clear_t1(wc);
		__t1_set_reg(wc,0x0,0x0);

		spin_unlock_irqrestore(&wc->lock, flags);

		/* Wait 100ms to give plenty of time for reset */
		wait(100);

		spin_lock_irqsave(&wc->lock, flags);

		__t1_set_reg(wc,0x70,(3<<1));
		__t1_set_reg(wc,0x1,(1<<1));
		__t1_set_reg(wc,0x2,(1<<0)|(1<<1));
		__t1_set_reg(wc,0x4f,(1<<0)|(1<<4));
		__t1_set_reg(wc,0x3,(1<<3));
		__t1_set_reg(wc,0x4,(1<<5)|(1<<6));
		__t1_set_reg(wc,0x5,(1<<4));
		__t1_set_reg(wc,0x6,(1<<7));
		__t1_set_reg(wc,0x7,(1<<2));
		__t1_set_reg(wc,0x40,0x0);
		__t1_set_reg(wc,0x79,(1<<3)|(1<<4));
		__t1_set_reg(wc,0x78,(1<<0));
		__t1_set_reg(wc,0x7b,(1<<1)|(1<<3));
		__t1_set_reg(wc,0x7a,0x0);
		__t1_set_reg(wc,0x71, 0x5);

		wc->span.flags |= ZT_FLAG_RUNNING;
	}

	__t1_set_reg(wc,0x70,(1<<2) | (wc->sync ? (1<<1) : 0));
	__t1_set_reg(wc,0x4,(lc & ZT_CONFIG_ESF ? (1<<6) : 0) | (lc & ZT_CONFIG_B8ZS ? (1<<5) : 0) | (japan ? (1<<1) : 0));
	if (mode == TYPE_T1) {
		__t1_set_reg(wc,0x6,(lc & ZT_CONFIG_B8ZS ? (1<<7) : (lc & ZT_CONFIG_ESF ? 0 : (1<<6))));
		__t1_set_reg(wc,0x7,(lc & ZT_CONFIG_ESF ? (1<<2) : 0));
		__t1_set_reg(wc,0xc1,(lc & ZT_CONFIG_ESF ? 0 : 0x1c));
	}
	if (japan) {
		__t1_set_reg(wc,0x4,(1<<1));
		__t1_set_reg(wc,0x5,(1<<4) | (1<<7));
	}
	coding = lc & ZT_CONFIG_ESF ? "ESF" : "SF";
	framing = lc & ZT_CONFIG_B8ZS ? "B8ZS" : "AMI";

	printk("Using %s/%s coding/framing\n", coding, framing);
	spin_unlock_irqrestore(&wc->lock, flags);
}

static void t1xxp_e1_framer_start(struct t1xxp *wc)
{
	char *coding, *framing, *crcing;
	int alreadyrunning = wc->span.flags & ZT_FLAG_RUNNING;
	unsigned long flags;
	int lc = wc->span.lineconfig;

	spin_lock_irqsave(&wc->lock, flags);

	if (!alreadyrunning) {
		__t1xxp_set_clear_e1(wc);
		__t1_set_reg(wc,0x0,(1<<1));

		spin_unlock_irqrestore(&wc->lock, flags);

		/* Wait 100ms to give plenty of time for reset */
		wait(100);

		spin_lock_irqsave(&wc->lock, flags);

		__t1_set_reg(wc,0x73, 1);
		__t1_set_reg(wc,0x70,(3<<1));
		__t1_set_reg(wc,0x1,(1<<1));
		__t1_set_reg(wc,0x2,(1<<0)|(1<<1));
		__t1_set_reg(wc,0x4f,(1<<0)|(1<<4));
		__t1_set_reg(wc,0x35,(1<<0)|(1<<2));
		__t1_set_reg(wc,0x36,(1<<2));
		__t1_set_reg(wc,0x33,(1<<3)|(1<<5)|(1<<6));
		__t1_set_reg(wc,0x34,(1<<0));
		__t1_set_reg(wc,0x40,(1<<1)|(1<<2));
		__t1_set_reg(wc,0xd0,(1<<0)|(1<<1)|(1<<3)|(1<<4));
		__t1_set_reg(wc,0xd1,(1<<0)|(1<<1)|(1<<2)|(1<<3)|(1<<4)|(1<<6));
		__t1_set_reg(wc,0x78,(1<<0)|(1<<5));
		__t1_set_reg(wc,0x79,(1<<4)|(1<<7));
		__t1_set_reg(wc,0x7b,(1<<0)|(1<<1)|(1<<2)|(1<<3));
		__t1_set_reg(wc,0x71, 0x5);

		wc->span.flags |= ZT_FLAG_RUNNING;
	}

	__t1_set_reg(wc,0x70,(1<<2) | (wc->sync ? (1<<1) : 0));
	__t1_set_reg(wc,0x33,(lc & ZT_CONFIG_CRC4 ? (1<<3) : 0) | (lc & ZT_CONFIG_HDB3 ? (1<<5) : 0) | (lc & ZT_CONFIG_CCS ? (1<<6) : 0));
	__t1_set_reg(wc,0x35,(lc & ZT_CONFIG_CRC4 ? (1<<0) : 0) | (lc & ZT_CONFIG_HDB3 ? (1<<2) : 0));
	__t1_set_reg(wc,0x40,(lc & ZT_CONFIG_CCS ? (1<<1)|(1<<2) : 0));

	coding = lc & ZT_CONFIG_CCS ? "CCS" : "CAS";
	framing = lc & ZT_CONFIG_HDB3 ? "HDB3" : "AMI";
	crcing = lc & ZT_CONFIG_CRC4 ? "/CRC4" : "/NOCRC4";

	printk("Using %s/%s coding/framing%s 120 Ohms\n", coding, framing,crcing);
	spin_unlock_irqrestore(&wc->lock, flags);
}

static int t1xxp_framer_sanity_check(struct t1xxp *wc)
{
	int res;
	int chipid;
	unsigned long flags;
	int x;

	/* Sanity check */
	spin_lock_irqsave(&wc->lock, flags);
	for (x=0x0;x<192;x++)
		__t1_set_reg(wc, x, 0);
	res = __t1_get_reg(wc, 0x0f);
	res = __t1_get_reg(wc, 0x0f);
	chipid = ((res & 0x80) >> 5) | ((res & 0x30) >> 4);
	wc->ise1 = mode;
	spin_unlock_irqrestore(&wc->lock, flags);

	printk("Framer: %s, Revision: %d (%s)\n", chips[chipid], res & 0xf, wc->ise1 ? "E1" : "T1");
	return 0;
}

static int t1xxp_framer_hard_reset(struct t1xxp *wc)
{
	int x;
	unsigned long flags;

	spin_lock_irqsave(&wc->lock, flags);
	/* Initialize all registers to 0 */
	for (x=0x0;x<192;x++)
		__t1_set_reg(wc, x, 0);

	spin_unlock_irqrestore(&wc->lock, flags);
	return 0;
}

static int t1xxp_rbsbits(struct zt_chan *chan, int bits)
{
	struct t1xxp *wc = chan->pvt;
	u_char m,c;
	int n,b;
	int all, byte;
	int location = 0;
	unsigned long flags;
#if 0
	printk("Setting bits to %d on channel %s\n", bits, chan->name);
#endif	
	all = 0;
	n = chan->chanpos - 1;
	if (wc->ise1) {
		if (chan->chanpos == 16) return 0;
		if (chan->chanpos > 16) {
			n--;
		}
		location++;
	}
	location += 0x50 + n/2;
	b = (n / 8);
	m = 1 << (n & 7);
	c = wc->txsigs[b];
	c &= ~m;
	if (bits & ZT_ABIT) c |= m;
	if (c & m)
		all |= (1<<3);
	wc->txsigs[b] = c;
	b += 3;
	c = wc->txsigs[b];
	c &= ~m;
	if (bits & ZT_BBIT) c |= m;
	if (c & m)
		all |= (1<<2);
	wc->txsigs[b] = c;
	b += 3;
	c = wc->txsigs[b];
	c &= ~m;
	if (bits & ZT_CBIT) c |= m;
	if (c & m)
		all |= (1<<1);

	wc->txsigs[b] = c;
	b += 3;
	c = wc->txsigs[b];
	c &= ~m;
	if (bits & ZT_DBIT) c |= m;
	if (c & m)
		all |= (1<<0);

	wc->txsigs[b] = c;
	spin_lock_irqsave(&wc->lock, flags);	
	byte = __t1_get_reg(wc, location);
	byte &= 0xf0 >> ((n%2)*4);
	byte |= all << ((n%2)*4);
	__t1_set_reg(wc, location, byte);
	spin_unlock_irqrestore(&wc->lock, flags);
	return 0;
}

static int t1xxp_ioctl(struct zt_chan *chan, unsigned int cmd, unsigned long data)
{
	switch(cmd) {
	default:
		return -ENOTTY;
	}
}

static int t1xxp_startup(struct zt_span *span)
{
	struct t1xxp *wc = span->pvt;

	int i,alreadyrunning = span->flags & ZT_FLAG_RUNNING;

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++)
	{
		memset(wc->ec_chunk1[i],
			ZT_LIN2X(0,&span->chans[i]),ZT_CHUNKSIZE);
		memset(wc->ec_chunk2[i],
			ZT_LIN2X(0,&span->chans[i]),ZT_CHUNKSIZE);
	}

	/* Reset framer with proper parameters and start */
	if (wc->ise1)
		t1xxp_e1_framer_start(wc);
	else
		t1xxp_t1_framer_start(wc);
	printk("Calling startup (flags is %d)\n", span->flags);

	if (!alreadyrunning) {
		/* Only if we're not already going */
		t1xxp_enable_interrupts(wc);
		t1xxp_start_dma(wc);
		span->flags |= ZT_FLAG_RUNNING;
	}
	return 0;
}

static int t1xxp_shutdown(struct zt_span *span)
{
	struct t1xxp *wc = span->pvt;
	unsigned long flags;

	spin_lock_irqsave(&wc->lock, flags);
	__t1xxp_stop_dma(wc);
	__t1xxp_disable_interrupts(wc);
	span->flags &= ~ZT_FLAG_RUNNING;
	spin_unlock_irqrestore(&wc->lock, flags);

	t1xxp_framer_hard_reset(wc);
	return 0;
}

/* needs to be implemented one day */
static int t1xxp_maint(struct zt_span *span, int cmd)
{
	return 0;
}

static int t1xxp_chanconfig(struct zt_chan *chan, int sigtype)
{
	struct t1xxp *wc = chan->pvt;
	unsigned long flags;
	int alreadyrunning = chan->span->flags & ZT_FLAG_RUNNING;

	spin_lock_irqsave(&wc->lock, flags);

	if (alreadyrunning) {
		if (wc->ise1)
			__t1xxp_set_clear_e1(wc);
		else
			__t1xxp_set_clear_t1(wc);
	}

	spin_unlock_irqrestore(&wc->lock, flags);
	return 0;
}

static int t1xxp_spanconfig(struct zt_span *span, struct zt_lineconfig *lc)
{
	struct t1xxp *wc = span->pvt;
	span->lineconfig = lc->lineconfig;
	span->txlevel = lc->lbo;
	span->rxlevel = 0;
	/* Do we want to SYNC on receive or not */
	wc->sync = lc->sync;
	/* If already running, apply changes immediately */
	if (span->flags & ZT_FLAG_RUNNING)
		return t1xxp_startup(span);
	return 0;
}

static int t1xxp_software_init(struct t1xxp *wc)
{
	int x;
	/* Find position */
	for (x=0;x<WC_MAX_CARDS;x++) {
		if (!cards[x]) {
			cards[x] = wc;
			break;
		}
	}
	if (x >= WC_MAX_CARDS)
		return -1;
	wc->num = x;
	sprintf(wc->span.name, "WC%s/%d", wc->ise1 ? "E1" : "T1", wc->num);
	sprintf(wc->span.desc, "%s Card %d", wc->variety, wc->num);
	wc->span.spanconfig = t1xxp_spanconfig;
	wc->span.chanconfig = t1xxp_chanconfig;
	wc->span.startup = t1xxp_startup;
	wc->span.shutdown = t1xxp_shutdown;
	wc->span.rbsbits = t1xxp_rbsbits;
	wc->span.maint = t1xxp_maint;
	wc->span.open = t1xxp_open;
	wc->span.close = t1xxp_close;
	if (wc->ise1)
		wc->span.channels = 31;
	else
		wc->span.channels = 24;
	wc->span.chans = wc->chans;
	wc->span.flags = ZT_FLAG_RBS;
	if(wc->ise1) {
		wc->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_CCS | ZT_CONFIG_HDB3 | ZT_CONFIG_CRC4;
	} else {
		wc->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_B8ZS | ZT_CONFIG_D4 | ZT_CONFIG_ESF;
	}
	wc->span.ioctl = t1xxp_ioctl;
	wc->span.watchdog = t1xxp_watchdog;
	wc->span.pvt = wc;
	if (wc->ise1)
		wc->span.deflaw = ZT_LAW_ALAW;
	else
		wc->span.deflaw = ZT_LAW_MULAW;
	init_waitqueue_head(&wc->span.maintq);
	for (x=0;x<wc->span.channels;x++) {
		sprintf(wc->chans[x].name, "WC%s/%d/%d", wc->ise1 ? "E1" : "T1", wc->num, x + 1);
		wc->chans[x].sigcap = ZT_SIG_EM | ZT_SIG_CLEAR | ZT_SIG_EM_E1 | 
				      ZT_SIG_FXSLS | ZT_SIG_FXSGS | 
				      ZT_SIG_FXSKS | ZT_SIG_FXOLS | ZT_SIG_DACS_RBS |
				      ZT_SIG_FXOGS | ZT_SIG_FXOKS | ZT_SIG_CAS | ZT_SIG_SF;
		wc->chans[x].pvt = wc;
		wc->chans[x].chanpos = x + 1;
	}
	if (zt_register(&wc->span, 0)) {
		printk("Unable to register span with zaptel\n");
		return -1;
	}
	return 0;
}

static inline void set_dma_memory(struct t1xxp *wc)
{
	/* Setup DMA Addresses */
	/* Start at writedma */
	outl(wc->writedma,                    wc->ioaddr + WC_DMAWS);		/* Write start */
	/* First frame */
	outl(wc->writedma + ZT_CHUNKSIZE * 32 - 4, wc->ioaddr + WC_DMAWI);		/* Middle (interrupt) */
	/* Second frame */
	outl(wc->writedma + ZT_CHUNKSIZE * 32 * 2 - 4, wc->ioaddr + WC_DMAWE);			/* End */
	
	outl(wc->readdma,                    	 wc->ioaddr + WC_DMARS);	/* Read start */
	/* First frame */
	outl(wc->readdma + ZT_CHUNKSIZE * 32 - 4, 	 wc->ioaddr + WC_DMARI);	/* Middle (interrupt) */
	/* Second frame */
	outl(wc->readdma + ZT_CHUNKSIZE * 32 * 2 - 4, wc->ioaddr + WC_DMARE);	/* End */
}

static inline void __handle_leds(struct t1xxp *wc)
{
	int oldreg;
	wc->blinktimer++;

	if (wc->span.alarms & (ZT_ALARM_RED | ZT_ALARM_BLUE)) {
		/* Red/Blue alarm */
#ifdef FANCY_ALARM
		if (wc->blinktimer == (altab[wc->alarmpos] >> 1)) {
			wc->ledtestreg = (wc->ledtestreg | BIT_LED1) & ~BIT_LED0;
			__t1_set_reg(wc, 0x73, wc->ledtestreg); 
		}
		if (wc->blinktimer == 0xf) {
			wc->ledtestreg = wc->ledtestreg & ~(BIT_LED0 | BIT_LED1);
			__t1_set_reg(wc, 0x73, wc->ledtestreg); 
			wc->blinktimer = -1;
			wc->alarmpos++;
			if (wc->alarmpos >= (sizeof(altab) / sizeof(altab[0])))
				wc->alarmpos = 0;
		}
#else
		if (wc->blinktimer == 160) {
			wc->ledtestreg = (wc->ledtestreg | BIT_LED1) & ~BIT_LED0;
			__t1_set_reg(wc, 0x73, wc->ledtestreg); 
		} else if (wc->blinktimer == 480) {
			wc->ledtestreg = wc->ledtestreg & ~(BIT_LED0 | BIT_LED1);
			__t1_set_reg(wc, 0x73, wc->ledtestreg); 
			wc->blinktimer = 0;
		}
#endif
	} else if (wc->span.alarms & ZT_ALARM_YELLOW) {
		/* Yellow Alarm */
		if (!(wc->blinktimer % 2)) 
			wc->ledtestreg = (wc->ledtestreg | BIT_LED1) & ~BIT_LED0;
		else
			wc->ledtestreg = (wc->ledtestreg | BIT_LED0) & ~BIT_LED1;
		__t1_set_reg(wc, 0x73, wc->ledtestreg); 
	} else {
		/* No Alarm */
		oldreg = wc->ledtestreg;
		if (wc->span.flags & ZT_FLAG_RUNNING)
			wc->ledtestreg = (wc->ledtestreg | BIT_LED0) & ~BIT_LED1;
		else
			wc->ledtestreg = wc->ledtestreg & ~(BIT_LED0 | BIT_LED1);
		if (oldreg != wc->ledtestreg)
			__t1_set_reg(wc, 0x73, wc->ledtestreg); 
	}
}

static void t1xxp_transmitprep(struct t1xxp *wc, int ints)
{
	volatile unsigned char *txbuf;
	int x,y;
	int chanmap;
	int pos;
	if (ints & 0x04 /* 0x01 */) {
		/* We just finished sending the first buffer, start filling it
		   now */
		txbuf = wc->writechunk;
	} else {
		/* Just finished sending second buffer, fill it now */
		txbuf = wc->writechunk + 32 * ZT_CHUNKSIZE;
	}
	zt_transmit(&wc->span);
	for (y=0;y<ZT_CHUNKSIZE;y++) {
		for (x=0;x<wc->span.channels;x++) {
			if (wc->ise1)
				chanmap = wc->chanmap[(x + 24 + 1) & 0x1f];
			else
				chanmap = wc->chanmap[((x + 18) % 24)+1];
			pos = y * 32 + chanmap;
#ifdef DATALINK
			if ((wc->ise1 && x >= 4) || (!wc->ise1 && x >= 3)) {
				txbuf[pos] = wc->tempo[chanmap];
				wc->tempo[chanmap] = wc->chans[x].writechunk[y];
			} else
#endif
			txbuf[pos] = wc->chans[x].writechunk[y];
		}
	}
}

static void t1xxp_receiveprep(struct t1xxp *wc, int ints)
{
	volatile unsigned char *rxbuf;
	volatile unsigned int *canary;
	int x;
	int y;
	unsigned int oldcan;
	if (ints & 0x04) {
		/* Just received first buffer */
		rxbuf = wc->readchunk;
		canary = (unsigned int *)(wc->readchunk + ZT_CHUNKSIZE * 64 - 4);
	} else {
		rxbuf = wc->readchunk + ZT_CHUNKSIZE * 32;
		canary = (unsigned int *)(wc->readchunk + ZT_CHUNKSIZE * 32 - 4);
	}
	oldcan = *canary;
	if (((oldcan & 0xffff0000) >> 16) != CANARY) {
		/* Check top part */
		if (debug) printk("Expecting top %04x, got %04x\n", CANARY, (oldcan & 0xffff0000) >> 16);
		wc->span.irqmisses++;
	} else if ((oldcan & 0xffff) != ((wc->canary - 1) & 0xffff)) {
		if (debug) printk("Expecting bottom %d, got %d\n", wc->canary - 1, oldcan & 0xffff);
		wc->span.irqmisses++;
	}
	for (y=0;y<ZT_CHUNKSIZE;y++) {
		for (x=0;x<wc->span.channels;x++) {
			/* XXX Optimize, remove * and + XXX */
			/* Must map received channels into appropriate data */
			int chanmap = wc->chanmap[x + 1];
			int pos = y * 32 + chanmap;
#ifdef DATALINK
			if ((wc->ise1 && x < 28) || (!wc->ise1 && x < 21)) {
				wc->chans[x].readchunk[y] = wc->temporx[chanmap];
				wc->temporx[chanmap] = rxbuf[pos];
			} else
#endif
			wc->chans[x].readchunk[y] = rxbuf[pos];
		}

#ifdef SANITY_CHECK
		if (!wc->clocktimeout && !wc->span.alarms) {
			unsigned char c = rxbuf[32 * y + wc->chanmap[0]];
			int miss = 0;
			if (wc->ise1) {
				if ((c & 0x7f) != 0x1b)
					miss = 1;
			} else {
				if (!y) {
					if (c != 0x7f)
						miss = 1;
				} else
					goto get_out_of_here;
			}
			if (miss) {
				if (wc->miss) {
					outb((1<<4), wc->ioaddr + WC_AUXD);
					set_dma_memory(wc);
					/* Wait 100ms to give plenty of time for reset */
					wait(100);
					outb(0x0, wc->ioaddr + WC_AUXD);
					wc->clocktimeout = 100;
					wc->miss = 0;
				} else {
					wc->miss = 1;
					wc->misslast = c;
				}
			} else
				wc->miss = 0;
get_out_of_here:;
		} else
#endif
			wc->miss = 0;
	}
	/* Store the next canary */
	canary = (unsigned int *)(rxbuf + ZT_CHUNKSIZE * 32 - 4);
	*canary = (wc->canary++) | (CANARY << 16);
	for (x=0;x<wc->span.channels;x++) {
		zt_ec_chunk(&wc->chans[x], wc->chans[x].readchunk, 
			wc->ec_chunk2[x]);
		memcpy(wc->ec_chunk2[x],wc->ec_chunk1[x],ZT_CHUNKSIZE);
		memcpy(wc->ec_chunk1[x],wc->chans[x].writechunk,ZT_CHUNKSIZE);
	}
	zt_receive(&wc->span);
}

static void __t1xxp_check_sigbits(struct t1xxp *wc, int i)
{
	int j, n;
	unsigned char rxc;
	unsigned char abits, bbits, cbits, dbits;

	if (wc->ise1) {
		/* if a E1, do the signalling */
		if (i < 12) {
			int bit[5];
			int checkbit = 0;
			n = (i % 3);
			for (j=0; j<10; j++) {
				i = (n * 10) + j;
				if (i >= 15) i++;
				if (!(wc->span.chans[i].sig & ZT_SIG_CLEAR))
					checkbit |= (1<<j);
			}
			for (j=0; j<5; j++) {
				if (checkbit & ((1<<(2*j)) || (1<<(2*j+1)))) {
					bit[j] = __t1_get_reg(wc, 0x61 + j + (5*n));
				}
			}
			for (j=0; j<10; j++) {
				if (!(checkbit & (1<<j)))
					continue;
				i = (n * 10) + j;
				if (i >= 15) i++;
				rxc = 0;
				abits = bit[j/2] & (1 << (4*((j%2)+1) - 1));
				bbits = bit[j/2] & (1 << (4*((j%2)+1) - 2));
				cbits = bit[j/2] & (1 << (4*((j%2)+1) - 3));
				dbits = bit[j/2] & (1 << (4*((j%2)+1) - 4));
				if (abits) rxc |= ZT_ABIT;
				if (bbits) rxc |= ZT_BBIT;
				if (cbits) rxc |= ZT_CBIT;
				if (dbits) rxc |= ZT_DBIT;
				if (wc->span.chans[i].rxsig != rxc)
					zt_rbsbits(&wc->span.chans[i], rxc);
			}
		}
	} else {
		/* if a T1, do the signalling */
		if (i < 12) {
			int bit[4];
			int checkbit = 0;
			n = (i % 3);
			for (j=0; j<8; j++) {
				i = (n * 8) + j;
				if (!(wc->span.chans[i].sig & ZT_SIG_CLEAR)) {
					checkbit |= (1<<j);
				}
			}
			for (j=0; j<4; j++) {
				if (checkbit & ((1<<(2*j)) || (1<<(2*j+1))))
					bit[j] = __t1_get_reg(wc, 0x60 + j + (4*n));
			}
			for (j=0; j<8; j++) {
				if (!(checkbit & (1<<j)))
					continue;
				/* Get channel number */
				i = (n * 8) + j;
				rxc = 0;
				abits = bit[j/2] & (1 << (4*((j%2)+1) - 1));
				bbits = bit[j/2] & (1 << (4*((j%2)+1) - 2));
				cbits = bit[j/2] & (1 << (4*((j%2)+1) - 3));
				dbits = bit[j/2] & (1 << (4*((j%2)+1) - 4));
				if (abits) rxc |= ZT_ABIT;
				if (bbits) rxc |= ZT_BBIT;
				if (cbits) rxc |= ZT_CBIT;
				if (dbits) rxc |= ZT_DBIT;
				if (wc->span.chans[i].rxsig != rxc)
					zt_rbsbits(&wc->span.chans[i], rxc);
			}
		}
	}
}

static void __t1xxp_check_alarms(struct t1xxp *wc)
{
	unsigned char c;
	int alarms;
	int x,j;
	__t1_set_reg(wc, 0x18, 0xff);
	c = __t1_get_reg(wc, 0x18);

	/* Assume no alarms */
	alarms = 0;

	/* And consider only carrier alarms */
	wc->span.alarms &= (ZT_ALARM_RED | ZT_ALARM_BLUE | ZT_ALARM_NOTOPEN);

	if (wc->span.lineconfig & ZT_CONFIG_NOTOPEN) {
		for (x=0,j=0;x < wc->span.channels;x++)
			if ((wc->chans[x].flags & ZT_FLAG_OPEN) ||
			    (wc->chans[x].flags & ZT_FLAG_NETDEV))
				j++;
		if (!j)
			alarms |= ZT_ALARM_NOTOPEN;
	}

	if (c & 0x1) { /* if red alarm */
		alarms |= ZT_ALARM_RED;
	}
	if (c & 0x2) { /* if blue alarm */
		alarms |= ZT_ALARM_BLUE;
	}

	/* Keep track of recovering */
	if ((!alarms) && wc->span.alarms)
		wc->alarmtimer = ZT_ALARMSETTLE_TIME;

	if (wc->span.alarms != alarms) {
		start_alarm(wc);
		if (!(alarms & (ZT_ALARM_RED | ZT_ALARM_BLUE | ZT_ALARM_LOOPBACK)) &&
		    wc->sync) {
			/* Use the recieve signalling */
			wc->span.syncsrc = wc->span.spanno;
		} else {
			wc->span.syncsrc = 0;
		}
	}
	if (wc->alarmtimer)
		alarms |= ZT_ALARM_RECOVER;
	if (c & 0x4)
		alarms |= ZT_ALARM_YELLOW;

	wc->span.alarms = alarms;

	zt_alarm_notify(&wc->span);
}

static void __t1xxp_do_counters(struct t1xxp *wc)
{
	if (wc->alarmtimer) {
		if (!--wc->alarmtimer) {
			wc->span.alarms &= ~(ZT_ALARM_RECOVER);
			zt_alarm_notify(&wc->span);
		}
	}
}

ZAP_IRQ_HANDLER(t1xxp_interrupt)
{
	struct t1xxp *wc = dev_id;
	unsigned char ints;
	unsigned long flags;
	int x;

	ints = inb(wc->ioaddr + WC_INTSTAT);
	outb(ints, wc->ioaddr + WC_INTSTAT);

	if (!ints)
#ifdef LINUX26
		return IRQ_NONE;
#else
		return;
#endif		

	if (!wc->intcount) {
		if (debug) printk("Got interrupt: 0x%04x\n", ints);
	}
	wc->intcount++;

	if (wc->clocktimeout)
		wc->clocktimeout--;

	if (ints & 0x0f) {
		t1xxp_receiveprep(wc, ints);
		t1xxp_transmitprep(wc, ints);
	}
	spin_lock_irqsave(&wc->lock, flags);

	__handle_leds(wc);

	/* Count down timers */
	__t1xxp_do_counters(wc);

	/* Do some things that we don't have to do very often */
	x = wc->intcount & 15 /* 63 */;

	__t1xxp_check_sigbits(wc, x);
	if (x == 4) {
		/* Check alarms 1/4 as frequently */
		if (!(wc->intcount & 0x30))
			__t1xxp_check_alarms(wc);
	}

	if (!(wc->intcount % 1000)) {/* even second boundary */
		/* add this second's BPV count to total one */
		wc->span.bpvcount += __t1_get_reg(wc,0x43) + (__t1_get_reg(wc,0x42) << 8);
		if (wc->span.lineconfig & ZT_CONFIG_CRC4) {
			wc->span.crc4count += __t1_get_reg(wc,0x45) + (__t1_get_reg(wc,0x44) << 8);
			wc->span.ebitcount += __t1_get_reg(wc,0x49) + (__t1_get_reg(wc,0x48) << 8);
		}
		wc->span.fascount += __t1_get_reg(wc,0x47) + (__t1_get_reg(wc,0x46) << 8);
	}
	
	spin_unlock_irqrestore(&wc->lock, flags);

	if (ints & 0x10) {
		printk("PCI Master abort\n");
		__t1xxp_stop_dma(wc);
	}

	if (ints & 0x20)
		printk("PCI Target abort\n");

#ifdef LINUX26
	return IRQ_RETVAL(1);
#endif		
}

static int t1xxp_hardware_init(struct t1xxp *wc)
{
	/* Hardware PCI stuff */
	/* Reset chip and registers */
	outb(DELAY | 0x0e, wc->ioaddr + WC_CNTL);
	/* Set all outputs to 0 */
	outb(0x00, wc->ioaddr + WC_AUXD);
	outb(0xff, wc->ioaddr + WC_AUXC);
	/* Configure the serial port: double clock, 20ns width, no inversion,
	   MSB first */
	outb((3<<6) | (1<<3) | (3<<1) | (1<<0), wc->ioaddr + WC_SERC);
	outb(0x00, wc->ioaddr + 0x2b); //AUXFUNC

	/* Internally delay FSC by one */
	outb(0x1, wc->ioaddr + WC_FSCDELAY);

	/* Back to normal, with automatic DMA wrap around */
	outb(DELAY | 0x01, wc->ioaddr + WC_CNTL);
	
	/* Make sure serial port and DMA are out of reset */
	outb(inb(wc->ioaddr + WC_CNTL) & 0xf9, WC_CNTL);

	/* Setup DMA Addresses */
	set_dma_memory(wc);	
	
	if (debug) printk("Setting up DMA (write/read = %08lx/%08lx)\n", (long)wc->writedma, (long)wc->readdma);

	/* Sanity check also determines e1 or t1 */
	if (t1xxp_framer_sanity_check(wc))
		return -1;
	if (wc->ise1)
		wc->chanmap = chanmap_e1;
	else
		wc->chanmap = chanmap_t1;

	wc->clocktimeout = 100;
	/* Reset the framer and report */
	t1xxp_framer_hard_reset(wc);
	start_alarm(wc);
	return 0;

}

static int __devinit t1xxp_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int res;
	struct t1xxp *wc;
	unsigned int *canary;
	
	if (pci_enable_device(pdev)) {
		res = -EIO;
	} else {
		wc = kmalloc(sizeof(struct t1xxp), GFP_KERNEL);
		if (wc) {
			memset(wc, 0x0, sizeof(struct t1xxp));
			spin_lock_init(&wc->lock);
			wc->ioaddr = pci_resource_start(pdev, 0);
			wc->dev = pdev;

			wc->writechunk = 
				/* 32 channels, Double-buffer, Read/Write */
				(unsigned char *)pci_alloc_consistent(pdev, ZT_MAX_CHUNKSIZE * 32 * 2 * 2, &wc->writedma);
			if (!wc->writechunk) {
				printk("raven: Unable to allocate DMA-able memory\n");
				return -ENOMEM;
			}

			/* Read is after the whole write piece (in bytes) */
			wc->readchunk = wc->writechunk + ZT_CHUNKSIZE * 32 * 2;

			/* Same thing...  */
			wc->readdma = wc->writedma + ZT_CHUNKSIZE * 32 * 2;

			/* Initialize Write/Buffers to all blank data */
			memset((void *)wc->writechunk,0x00,ZT_MAX_CHUNKSIZE * 2 * 2 * 32);
			/* Initialize canary */
			canary = (unsigned int *)(wc->readchunk + ZT_CHUNKSIZE * 64 - 4);
			*canary = (CANARY << 16) | (0xffff);

			/* Enable bus mastering */
			pci_set_master(pdev);

			/* Keep track of which device we are */
			pci_set_drvdata(pdev, wc);

			if (request_irq(pdev->irq, t1xxp_interrupt, ZAP_IRQ_SHARED_DISABLED, "raven", wc)) {
				printk("raven: Unable to request IRQ %d\n", pdev->irq);
				kfree(wc);
				return -EIO;
			}
			/* Initialize hardware */
			t1xxp_hardware_init(wc);

			/* We now know which version of card we have */
			if (wc->ise1)
				wc->variety = "PhonicEQ Inc. E100P RAVEN E1/PRA";
			else
				wc->variety = "PhonicEQ Inc. T100P RAVEN T1/PRI";

			/* Misc. software stuff */
			t1xxp_software_init(wc);

			printk("Found a card: %s\n", wc->variety);
			res = 0;
		} else
			res = -ENOMEM;
	}
	return res;
}

static void t1xxp_stop_stuff(struct t1xxp *wc)
{
	/* Reset the framer */
	t1xxp_framer_hard_reset(wc);
}

static void __devexit t1xxp_remove_one(struct pci_dev *pdev)
{
	struct t1xxp *wc = pci_get_drvdata(pdev);
	if (wc) {

		/* Stop any DMA */
		__t1xxp_stop_dma(wc);

		/* In case hardware is still there */
		__t1xxp_disable_interrupts(wc);
		
		t1xxp_stop_stuff(wc);

		/* Immediately free resources */
		pci_free_consistent(pdev, ZT_MAX_CHUNKSIZE * 2 * 2 * 32 * 4, (void *)wc->writechunk, wc->writedma);
		free_irq(pdev->irq, wc);

		/* Reset PCI chip and registers */
		outb(DELAY | 0x0e, wc->ioaddr + WC_CNTL);

		/* Release span, possibly delayed */
		if (!wc->usecount)
			t1xxp_release(wc);
		else
			wc->dead = 1;
	}
}

static struct pci_device_id t1xxp_pci_tbl[] = {
	{ 0xe159, 0x0001, 0x2801, 0x0003, 0, 0, (unsigned long) "PhonicEQ Inc. E100P RAVEN T1/PRI E1/PRI Board" },
	{ 0 }
};

MODULE_DEVICE_TABLE(pci,t1xxp_pci_tbl);

static struct pci_driver t1xxp_driver = {
	name: 	"raven",
	probe: 	t1xxp_init_one,
#ifdef LINUX26
	remove:	__devexit_p(t1xxp_remove_one),
#else
	remove:	t1xxp_remove_one,
#endif
	suspend: NULL,
	resume:	NULL,
	id_table: t1xxp_pci_tbl,
};

static int __init t1xxp_init(void)
{
	int res;
	res = zap_pci_module(&t1xxp_driver);
	if (res)
		return -ENODEV;
	return 0;
}

static void __exit t1xxp_cleanup(void)
{
	pci_unregister_driver(&t1xxp_driver);
}

#ifdef LINUX26
module_param(debug, int, 0600);
module_param(mode, int, 0600);
module_param(japan, int, 0600);
#else
MODULE_PARM(debug, "i");
MODULE_PARM(mode, "i");
MODULE_PARM(japan, "i");
#endif
MODULE_DESCRIPTION("T100P/E100P RAVEN T1/E1/PRA Zaptel Driver");
MODULE_AUTHOR("Marcin Pycko <mpycko@phoniceq.com>");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

module_init(t1xxp_init);
module_exit(t1xxp_cleanup);
